// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import PaillierKey from './paillierKey';

export default class PaillierScalar {
    /**
     * @param {bigInt.BigInteger} c
     * @param {PaillierKey} key
     */
    constructor(c, key) {
        this.c = c;
        this.key = key;
    }

    /** @param {(PaillierScalar|bigInt.BigInteger)} scalar */
    plus(scalar) {
        return this.add(scalar);
    }

    /** @param {(PaillierScalar|bigInt.BigInteger)} scalar */
    add(scalar) {
        return scalar instanceof PaillierScalar ? addScalar(this, scalar) : addValue(this, scalar);
    }

    negate() {
        return negateScalar(this);
    }

    /** @param {(PaillierScalar|bigInt.BigInteger)} scalar */
    subtract(scalar) {
        return this.minus(scalar);
    }

    /** @param {(PaillierScalar|bigInt.BigInteger)} scalar */
    minus(scalar) {
        return scalar instanceof PaillierScalar ? minusScalar(this, scalar) : minusValue(this, scalar);
    }

    /** @param {bigInt.BigInteger} scalar */
    multiply(scalar) {
        return this.times(scalar);
    }

    /** @param {bigInt.BigInteger} scalar */
    times(scalar) {
        return timesValue(this, scalar)
    }

    /** @param {(PaillierScalar|bigInt.BigInteger)} scalar */
    equals(scalar) {
        return scalar instanceof PaillierScalar ? this.c.equals(scalar.c) : this.c.equals(scalar);
    }
}

/**
 * @param {PaillierScalar} c1
 * @param {bigInt.BigInteger} m2
 */
function addValue(c1, m2) {
    return new PaillierScalar(c1.key.g.modPow(m2, c1.key.nsquare).times(c1.c).mod(c1.key.nsquare), c1.key);
}

/**
 * @param {PaillierScalar} c1
 * @param {PaillierScalar} c2
 */
function addScalar(c1, c2) {
    if (c1.key !== c2.key)
        throw new Error("The values does not share the same key");

    return new PaillierScalar(c1.c.times(c2.c).mod(c1.key.nsquare), c1.key);
}

/** @param {PaillierScalar} c1 */
function negateScalar(c1) {
    return new PaillierScalar(c1.c.modInv(c1.key.nsquare), c1.key);
}

/**
 * @param {PaillierScalar} c1
 * @param {bigInt.BigInteger} m2
 */
function minusValue(c1, m2) {
    var neg = c1.key.nsquare.minus(m2);
    return new PaillierScalar(c1.key.g.modPow(neg, c1.key.nsquare).times(c1.c).mod(c1.key.nsquare), c1.key);
}

/**
 * @param {PaillierScalar} c1
 * @param {PaillierScalar} c2
 */
function minusScalar(c1, c2) {
    if (c1.key !== c2.key)
        throw new Error("The values does not share the same key");

    return new PaillierScalar(c1.c.times(c2.c.modInv(c1.key.nsquare)).mod(c1.key.nsquare), c1.key);
}

/**
 * @param {PaillierScalar} c1
 * @param {bigInt.BigInteger} m2
 */
function timesValue(c1, m2) {
    return new PaillierScalar(c1.c.modPow(m2, c1.key.nsquare), c1.key);
}
